//
// File:        Extendable.java
// Package:     gov.llnl.babel.symbols
// Revision:    @(#) $Id: Extendable.java 6212 2007-11-01 00:07:17Z epperly $
// Description: sidl base symbol for both classes and interfaces
//
// Copyright (c) 2000-2003, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel.symbols;

import gov.llnl.babel.Context;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.Utilities;
import gov.llnl.babel.symbols.Assertion;
import gov.llnl.babel.symbols.AssertionException;
import gov.llnl.babel.symbols.Comment;
import gov.llnl.babel.symbols.Interface;
import gov.llnl.babel.symbols.Metadata;
import gov.llnl.babel.symbols.Method;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.SymbolID;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Class <code>Extendable</code> is a base class for SIDL symbols of type
 * class and interface.  It brings together common methods and implementation
 * for both final classes.  Class and interfaces have a set of methods and
 * interface inheritance in common.  Classes have the additional property
 * that they can be extended by another class; that functionality is not
 * defined here.  Many of the member functions take a boolean argument that
 * selects whether the method refers to this particular extendable only or
 * to this extendable and all of its parents in the SIDL inheritance system.
 * Constraints on the validity of methods added to this object must be checked
 * by the parser.
 *
 * Key design goals include providing mechanisms to:
 * - ensure method lookups in O(1) by both long and short names;
 * - ensure fast return of abstract, static, non-static, local, and both
 *   local and parent (or all) methods;
 * - ensure original ordering of methods preserved when returning lists;
 */
public abstract class Extendable extends Symbol {
  private List d_all_abstract_methods;
  private Set   d_all_basicarrays;
  private Map   d_all_interfaces;
  private List d_all_methods;
  private List d_all_methods_with_nonblocking;
  private Map   d_all_methods_long;
  private Map   d_all_methods_short;
  private Set   d_all_references;
  private List d_all_nonstatic_methods;
  private List d_all_static_methods;

  private Map   d_local_interfaces;
  private List d_local_methods;
  private List d_local_methods_with_nonblocking;
  private Map   d_local_methods_long;
  private Map   d_local_methods_short;
  private List d_local_nonstatic_methods;
  private List d_local_static_methods;

  private Map   d_renamed_parent_methods; //For from clauses, maps from
                                              //new Method to old Method
  private Map   d_renamed_method_symbolID;//For from clauses, maps from  
                                              //old Method to it's enclosing 
                                              //SymbolID
  private Map  d_old_name_new_method;  //Maps from the fully qualified
                                            //name of the old method, to the
                                            //complete new Method.
   
  private List d_invariants;		// Local invariants ONLY


  /**
   * Create an empty <code>Extendable</code> object that will be constructed
   * by calls to other member functions.
   *
   * @param  id       The symbol id of the new object
   * @param  type     The value of the type of the new object
   * @param  comment  The comment associated with the object
   */
  public Extendable(SymbolID id, 
                    int type, 
                    Comment comment, 
                    Context context) {
    super(id, type, comment, context);
    d_all_abstract_methods    = new ArrayList();
    d_all_basicarrays         = new HashSet();
    d_all_interfaces          = new HashMap();
    d_all_methods             = new ArrayList();
    d_all_methods_long        = new HashMap();
    d_all_methods_short       = new HashMap();
    d_all_references          = new HashSet();
    d_all_nonstatic_methods   = new ArrayList();
    d_all_static_methods      = new ArrayList();

    d_local_interfaces        = new HashMap();
    d_local_methods           = new ArrayList();
    d_local_methods_long      = new HashMap();
    d_local_methods_short     = new HashMap();
    d_local_nonstatic_methods = new ArrayList();
    d_local_static_methods    = new ArrayList();

    d_renamed_parent_methods = new HashMap();
    d_renamed_method_symbolID = new HashMap();
    d_old_name_new_method = new HashMap();

    d_invariants              = new ArrayList();

    d_all_references.add(id);
  }

  /**
   * Create an empty <code>Extendable</code> object that will be constructed
   * by calls to other member functions.
   *
   * @param  id       The symbol id of the new object
   * @param  type     The value of the type of the new object
   * @param  comment  The comment associated with the object
   * @param  m        Object metadata
   */
  public Extendable(SymbolID id, int type, Comment comment, 
                    Metadata m, Context context) {
    super(id, type, comment, m, context);
    d_all_abstract_methods    = new ArrayList();
    d_all_basicarrays         = new HashSet();
    d_all_interfaces          = new HashMap();
    d_all_methods             = new ArrayList();
    d_all_methods_long        = new HashMap();
    d_all_methods_short       = new HashMap();
    d_all_references          = new HashSet();
    d_all_nonstatic_methods   = new ArrayList();
    d_all_static_methods      = new ArrayList();

    d_local_interfaces        = new HashMap();
    d_local_methods           = new ArrayList();
    d_local_methods_long      = new HashMap();
    d_local_methods_short     = new HashMap();
    d_local_nonstatic_methods = new ArrayList();
    d_local_static_methods    = new ArrayList();

    d_renamed_parent_methods = new HashMap();
    d_renamed_method_symbolID = new HashMap();
    d_old_name_new_method = new HashMap();

    d_invariants              = new ArrayList();

    d_all_references.add(id);
  }

  /**
   * Return TRUE if this object contains any abstract methods; otherwise, 
   * return FALSE.  A class is abstract if and only if it has any abstract 
   * methods.  An interface must always be abstract, even if it contains no 
   * methods.
   */
  abstract public boolean isAbstract();

  /**
   * Return TRUE if this object represents an interface, FALSE if it is a 
   * class.
   */
  abstract public boolean isInterface();

  /**
   * Add the specified new method to this object.  No checking is done whether 
   * this method is valid for this particular extendable.  The new method will 
   * over-write any existing method unless the new method is abstract.  The
   * references and arrays will also be cached in this object.
   *
   * @param  method  The method to be added to this object
   */
  public void addMethod(Method method) {
    checkFrozen();
    addToMethodLists(method, false);
    d_all_references.addAll(method.getSymbolReferences());
    d_all_basicarrays.addAll(method.getBasicArrays());
  }

  /** 
   * For From clauses:Add the specified "new" method to the normal method lists, and the
   * "old" method to the d_renamed_parent_methods HashMap.  newM and oldM
   * are assumed to have the same signatures and only differ by name.
   */
  public void addRenamedMethod(Method newM, Method oldM, SymbolID old_sid ) {
    checkFrozen();
    removeFromMethodLists(oldM);  //Remove the renamed method if it exists.
    addMethod(newM);
    d_renamed_parent_methods.put(newM, oldM);
    d_renamed_method_symbolID.put(oldM, old_sid);
    d_old_name_new_method.put(old_sid.getFullName()+"."+oldM.getLongMethodName(), newM);
  }
  

  /** 
   * For From clauses:Find the Parent Method Object for a renamed method.
   */
  public Method getRenamedMethod(Method newM) {
    return (Method)d_renamed_parent_methods.get(newM);
  }

  /** 
   * For From clauses:Get all the orginal Method Objects of all the renamed methods
   */
  public Set getRenamedMethods() {
    return d_renamed_method_symbolID.keySet();
  }

  /** 
   * For From clauses:Get all the new method objects of all the renamed methods
   */
  public Set getNewMethods() {
    return d_renamed_parent_methods.keySet();
  }


  /** 
   * For From clauses:Check a parent has a different name for this method.
   */
  public boolean methodWasRenamed(Method newM) {
    return d_renamed_parent_methods.get(newM) != null;
  }

  /** 
   * For From clauses: Get the old old method's enclosing symbolID. 
   */
  public SymbolID getRenamedMethodSymbolID(Method oldM) {
    return (SymbolID) d_renamed_method_symbolID.get(oldM);
  }

  /** 
   * For From clauses: Take the FQN of the renamed method and return the new
   * Method.
   */
  public Method getNewMethod(String oldM_name) {
    return (Method) d_old_name_new_method.get(oldM_name);
  }

  /**
   * Return the number of methods associated with this extendable.
   *
   * @param  all  If TRUE, then return the number of local and parent methods; 
   *              otherwise, return the number of local methods only.
   */
  public int getNumberOfMethods(boolean all) {
    return all ? d_all_methods.size() : d_local_methods.size();
  }

  /**
   * Return the methods in this interface as a <code>Collection</code>.
   * Each element in the collection is of type <code>Method</code>.
   *
   * @param  all  If TRUE, then return local and parent methods; otherwise,
   *              return only local methods.
   */
  public List getMethods(boolean all) {
    return all ? d_all_methods : d_local_methods;
  }

  /**
   * Return the methods in this interface as a <code>Collection</code>.
   * Each element in the collection is of type <code>Method</code>.
   * This list includes methods NONBLOCKING_SEND and NONBLOCKING_RECV
   * methods spawned in response to NONBLOCKING methods.
   *
   * @param  all  If TRUE, then return local and parent methods; otherwise,
   *              return only local methods.
   */
  public List getMethodsWithNonblocking(boolean all) throws CodeGenerationException {
    if(d_all_methods_with_nonblocking == null) {
      d_all_methods_with_nonblocking = new ArrayList();
      d_local_methods_with_nonblocking = new ArrayList();
      for(Iterator allMethods = d_all_nonstatic_methods.iterator(); allMethods.hasNext();) {
        Method method = (Method) allMethods.next();

        findExistingEntry(d_all_methods_with_nonblocking, method, true);
        if(method.getCommunicationModifier() == Method.NONBLOCKING) {
          Method send = method.spawnNonblockingSend();
          Method recv = method.spawnNonblockingRecv();
          findExistingEntry(d_all_methods_with_nonblocking, send, true);
          findExistingEntry(d_all_methods_with_nonblocking, recv, true);
        }
      }
      for(Iterator localMethods = d_local_nonstatic_methods.iterator(); localMethods.hasNext();) {
        Method method = (Method) localMethods.next();
        
        findExistingEntry(d_local_methods_with_nonblocking, method, true);
        if(method.getCommunicationModifier() == Method.NONBLOCKING) {
          Method send = method.spawnNonblockingSend();
          Method recv = method.spawnNonblockingRecv();
          findExistingEntry(d_local_methods_with_nonblocking, send, true);
          findExistingEntry(d_local_methods_with_nonblocking, recv, true);
        }
      }
    }
    return all ? d_all_methods_with_nonblocking 
      : d_local_methods_with_nonblocking;
  }


  /**
   * Return the <code>Collection</code> non-static methods in this interface.
   * Each element in the collection is of type <code>Method</code>.
   *
   * @param  all  If TRUE, then return local and parent non-static methods; 
   *              otherwise, return only local non-static methods.
   */
  public List getNonstaticMethods(boolean all) {
    return all ? d_all_nonstatic_methods
      : d_local_nonstatic_methods;
  }

  /**
   * Return the static methods in this interface as a <code>Collection</code>.
   * Each element in the collection is of type <code>Method</code>.
   *
   * @param  all  If TRUE, then return local and parent static methods; 
   *              otherwise, return only local static methods.
   */
  public List getStaticMethods(boolean all) {
    return all ? d_all_static_methods
      : d_local_static_methods;
  }

  /**
   * Return the abstract methods for this class or interface, which
   * includes all parent classes and interfaces.  Each element in the
   * collection is of type <code>Method</code>.
   */
  public List getAbstractMethods() {
    return d_all_abstract_methods;
  }

  /**
   *  Return all abstract methods and all local methods (both abstract and concreate).
   *  This is a special case for Java abstract classes, although maybe it
   *  will be useful elsewhere.
   */
  public List getAbstractAndLocalMethods() {
    LinkedList ret = new LinkedList(d_local_methods);
    //Add all Abstracts that don't collide with locals
    for (Iterator abs = d_all_abstract_methods.iterator(); abs.hasNext(); ) {
      Method absM = (Method) abs.next();
      for (Iterator local = d_local_methods.iterator(); local.hasNext(); ) {
        Method locM = (Method) local.next();
        if(absM == locM)
          break;
        if(!local.hasNext())
          ret.add(absM);
      }
    }
    return ret;
  }
    

  /**
   * Return <code>true</code> if and only if the method given is locally
   * defined in this extendable.
   *
   * @param  m  The method of interest.
   * @return <code>true</code> if <code>m</code> is locally defined
   *         in this Extendable.
   */
  public boolean isLocal(Method m)
  {
    return (m != null) &&
      m.equals(d_local_methods_long.get(m.getLongMethodName()));
  }

  /**
   * Return the <code>Method</code> with the specified long method name.  
   * If there is none, return null.
   *
   * @param  name  The short method name for the method to be located.
   * @param  all   If TRUE then all local and parent methods are to 
   *               be searched; otherwise, only local methods are to
   *               be searched.
   */
  public Method lookupMethodByLongName(String name, boolean all) {
    return (Method) (all ? d_all_methods_long.get(name) 
                     : d_local_methods_long.get(name));
  }

  
  /**
   * Return the <code>Collection</code> of methods, each in <code>Method</code>
   * form, that are found to have the specified short method name, if any.  
   * If none are found, then return null.
   *
   * @param  name  The short method name for the method to be located.
   * @param  all   If TRUE then all local and parent methods are to 
   *               be searched; otherwise, only local methods are to
   *               be searched.
   */
  public Collection lookupMethodByShortName(String name, boolean all) {
    return protectCollection((Collection)
                             (all ? d_all_methods_short.get(name)
                              : d_local_methods_short.get(name)));
  }

  /**
   * Return TRUE if the specified method exists by long name; otherwise,
   * return FALSE.
   *
   * @param  name  The long method name for the method to be located.
   * @param  all   If TRUE then all local and parent methods are to 
   *               be searched; otherwise, only local methods are to
   *               be searched.
   */
  public boolean hasMethodByLongName(String name, boolean all) {
    return all ? d_all_methods_long.containsKey(name) 
      : d_local_methods_long.containsKey(name);
  }

  /**
   * Return TRUE if the specified method exists by short name; otherwise,
   * return FALSE.  Recall there may be multiple methods with the same short 
   * name but here we only care if there is at least one.
   *
   * @param  name  The short method name to be located
   * @param  all   If TRUE, then all local and parent methods are to be 
   *               searched; otherwise, only local methods are to be 
   *               searched
   */
  public boolean hasMethodByShortName(String name, boolean all) {
    return all ? d_all_methods_short.containsKey(name) 
      : d_local_methods_short.containsKey(name);
  }

  /**
   * Return TRUE if any of the methods are static methods; otherwise,
   * return FALSE.
   *
   * @param  all  If TRUE, then local and parent methods are to be searched;
   *              otherwise, only local methods are to be searched.
   */
  abstract public boolean hasStaticMethod(boolean all);

  /**
   * Return TRUE if any of the methods throws an exception; otherwise, return
   * FALSE.  
   *
   * @param  all  If TRUE, then local and parent methods are to be searched;
   *              otherwise, only local methods are to be searched.
   */
  public boolean hasExceptionThrowingMethod(boolean all) {
    boolean has_exception_throwing_method = false;
     
    for (Iterator i = getMethods(all).iterator(); i.hasNext(); ) {
      Method method = (Method) i.next();
      if (method.getThrows().size()>0) { 
        has_exception_throwing_method = true;
        break;
      }
    }
    return has_exception_throwing_method;
  }

  /**
   * Add a new parent interface to this object.  This method will be
   * implemented by the <code>Class</code> and <code>Interface</code>
   * subclasses.
   *
   * @param  parent  The parent interface to be added to this object.
   */
  abstract public void addParentInterface(Interface parent);

  /**
   * Return true if this extendable inherts from interfaces. 
   * BaseInterface does not count for this test.
   *
   */
  public boolean hasParentInterfaces() {
    return (d_all_interfaces.size() > 1);
  }

  /**
   * Return the parent interfaces in a <code>Collection</code>.  Each
   * member of the collection is an <code>Interface</code>.  
   *
   * @param  all  If TRUE, then all parents are to be returned; otherwise,
   *              only direct parents are to be returned.
   */
  public Collection getParentInterfaces(boolean all) {
    return all 
      ? d_all_interfaces.values() 
      : d_local_interfaces.values();
  }

  /**
   * Return the parent interfaces and/or classes in hierarchical order
   * from top-most down to direct parents in an <code>List</code>.  
   * Each member of the list is an <code>Extendable</code> and there
   * are NO duplicates.
   * Does a DFS to return ancestors in "reverse" topolgical order.
   * by Jim Leek
   */
  public List getAllParentsInOrder() throws CodeGenerationException {
    LinkedList ancestors = new LinkedList();
    HashSet visited = new HashSet();
    visited.add(this);  //We actually don't want this in the ancestor list.

    getAllParentsInOrderHelper(this, ancestors, visited);
    return ancestors;
  }

  /**
   * getAllParentsInOrderHelper is a recursive DFS call.  It is started
   * by getAllParentsInOrder.  It returns the post order list of ancestors.
   * (a "reverse" topological sort)
   */
  private void getAllParentsInOrderHelper(Extendable current, List ancestors, Set visited) 
    throws CodeGenerationException {

    Collection parents = Utilities.sort(current.getParents(false));

    for(Iterator pIter = parents.iterator(); pIter.hasNext();) {
      Extendable parent = (Extendable) pIter.next();
      if(!visited.contains(parent)) {
        getAllParentsInOrderHelper(parent, ancestors, visited);
      }
    }
    
    //Keep "this" out of the ancestor list
    if(!visited.contains(current)) {
      //This node sould never have been visited before, assert that.
      if(!ancestors.contains(current)) {
        ancestors.add(current); 
        visited.add(current);  //Finished visiting this node
      } else {
        throw new CodeGenerationException 
          ("Extendable: " + getFullName() + " Had ancestor " + 
           current.getFullName() + " appear in the ancestor list twice.");
      }
    }
  }

  /**
   * Return the parent interfaces and/or class in a <code>Collection</code>.
   * Each member of the collection is an <code>Extendable</code>.  No 
   * duplicate entries are returned in the collection.
   *
   * @param  all  If TRUE, all ancestors are to be returned; otherwise, only
   *              direct parents.  
   */
  public Collection getParents(boolean all) {
    return protectCollection(getParentInterfaces(all));
  }

  /**
   * Return TRUE if this object implements or extends the specified 
   * <code>Extendable</code> directly or indirectly.  Otherwise, return 
   * FALSE.
   *
   * @param  ext  The potential ancestor
   */
  public boolean hasAncestor(Extendable ext)
  {
    if (ext == null) return false;
    Collection parents = getParents(false);
    Iterator i = parents.iterator();
    while (i.hasNext()) {
      Extendable parent = (Extendable)i.next();
      if (ext == parent) return true;
    }
    i = parents.iterator();
    while (i.hasNext()) {
      Extendable parent = (Extendable)i.next();
      if (parent.hasAncestor(ext)) return true;
    }
    return false;
  }

  /**
   * Return TRUE if this class has the specified parent interface; otherwise,
   * return FALSE.  
   *
   * @param  id   The parent interface to be located
   * @param  all  Indicator of whether all parents (TRUE) or only direct
   *              parents (FALSE) are to be searched
   */
  public boolean hasParentInterface(SymbolID id, boolean all) {
    return all ? d_all_interfaces.containsKey(id)
      : d_local_interfaces.containsKey(id);
  }

  /**
   * Return Interface if this class has the specified parent interface; otherwise,
   * return FALSE.  
   *
   * @param  id   The parent interface to be located
   * @param  all  Indicator of whether all parents (TRUE) or only direct
   *              parents (FALSE) are to be searched
   */
  public Interface getParentInterface(SymbolID id, boolean all) {
    return (Interface) (all ? d_all_interfaces.get(id)
                        : d_local_interfaces.get(id));
  }

  /**
   * Return the <code>Set</code> of symbol references for this object, each
   * element of which is in <code>SymbolID</code> form.  These are
   * defined as all references for this object as well as its parents.
   * The set of references includes this symbol name.
   */
  public Set getSymbolReferences() {
    return d_all_references;
  }

  /**
   * Return the <code>Set</code> of all symbol references for this object, each
   * element of which is in <code>SymbolID</code> form.  Note the references
   * include those from this object's parents.
   */
  public Set getAllSymbolReferences() {
    return getSymbolReferences();
  }

  /**
   * Return the <code>Set</code> of basic array types for this object, each
   * element of the set being in <code>SymbolID</code> form.
   */
  public Set getBasicArrayRefs() {
    return d_all_basicarrays;
  }

  /**
   * Protected method called by parents to add their relevant information,
   * such as methods, interfaces, references, and arrays to this object.
   *
   * @param  ext  The parent whose methods and related information are to be 
   *              added
   */
  protected void addParentData(Extendable ext) {
    checkFrozen();
    /*
     * Add all parent methods to the list of all methods in this
     * class.  Do not override an existing method in the methods
     * list if the new method is abstract, since abstract methods
     * cannot override other methods.  If the method is abstract,
     * then add it to the list of abstract methods; otherwise, remove
     * it from the list.
     */
    for (Iterator m = ext.getMethods(true).iterator(); m.hasNext(); ) {
      Method method = (Method) m.next();
      //Only add the method if it hasn't already been renamed.
      if(!d_renamed_method_symbolID.containsKey(method)) {
        addToMethodLists(method, true);
      }
    }

    /*
     * Add the parent and its parents to the list of parents.  That is
     * a lot of parents, let me tell you.
     */
    if (ext.isInterface()) {
      d_all_interfaces.put(ext.getSymbolID(), ext);
      d_local_interfaces.put(ext.getSymbolID(), ext);
    }

    Collection parents = ext.getParentInterfaces(true);
    for (Iterator p = parents.iterator(); p.hasNext(); ) {
      Extendable e = (Extendable) p.next();
      d_all_interfaces.put(e.getSymbolID(), e);
    }

    d_all_references.addAll(ext.getSymbolReferences());
    d_all_basicarrays.addAll(ext.getBasicArrayRefs());
  }

  /**
   *  Get all methods with the same short name as the passed in method name.
   *  (This list includes the passed in method, so the collection has a
   *  minimum size of 1)
   *  This is necessary for F90 method overloading
   */
  public Collection getOverloadedMethodsByName(String shortName) {
    LinkedList ret = new LinkedList();
    for(Iterator i = d_all_methods.iterator(); i.hasNext();) {
      Method m = (Method) i.next();
      if(m.getShortMethodName().compareTo(shortName) == 0) {
        ret.add(m);
      }
    }
    return ret;
  }

  /**
   * Locate the corresponding <code>Method</code> in the given list of 
   * <code>Method</code> elements, removing and optionally replacing
   * it.
   *
   * Assumptions:  
   * 1) There is only one entry in the list for each long method name.
   *    (This is appropriate since the long method name is supposed to be 
   (    unique.)
   *
   * @param  list       List of methods to be searched.
   * @param  method     The method whose existing entry is to be located
   * @param  replaceIt  Indicator of whether or not the existing method is
   *                    to be replaced
   */
  private void findExistingEntry(List list, Method method, 
                                 boolean replaceIt)
  {
    String  name  = method.getLongMethodName();
    boolean found = false;
    Method  item  = null;

    for (Iterator iter = list.iterator(); iter.hasNext() && !found; )
      {
        item = (Method) iter.next();
        if (item.getLongMethodName().equals(name)) {
          found = true;
          iter.remove();
        }
      }
    if (replaceIt) {
      list.add(method);
    }
  }

  /**
   * Add the given method to the specified short-name method list.
   *
   * @param  list    The short-name method list
   * @param  method  The method to be added to the list
   */
  private void addToShortMethodList(Map list, Method method) {
    checkFrozen();
    String name = method.getShortMethodName();

    if (list.containsKey(name)) {
      findExistingEntry((List) list.get(name), method, true);
    } else {
      ArrayList sublist = new ArrayList();
      sublist.add(method);
      list.put(name, sublist);
    }
  }


  /**
   * Remove the given method from the specified short-name method list.
   *
   * @param  list    The short-name method list
   * @param  method  The method to be added to the list
   */
  private void removeFromShortMethodList(Map list, Method method) {
    String name = method.getShortMethodName();

    if (list.containsKey(name)) {
      findExistingEntry((List) list.get(name), method, false);
    } 
  }

  /**
   * Add the specified method to all and, if not isParent, local method lists
   * provided the method does not exist or is not abstract.  Note the latter 
   * case is used to overwrite existing methods.
   *
   * @param  method    The method to be added
   * @param  isParent  TRUE if method belongs to the parent; otherwise, FALSE.
   */
  private void addToMethodLists(Method method, boolean isParent) {
    String long_name = method.getLongMethodName();

    if (!(hasMethodByLongName(long_name, true) && method.isAbstract())) {
      boolean is_static  = method.isStatic();

      /* 
       * First process the lists of local and parent methods (i.e., the
       * "all" lists).
       */
      findExistingEntry(d_all_methods, method, true);
      d_all_methods_long.put(long_name, method);
      addToShortMethodList(d_all_methods_short, method);

      if (is_static) {
        findExistingEntry(d_all_static_methods, method, true);
      } else {
        findExistingEntry(d_all_nonstatic_methods, method, true);
      }
      if (method.isAbstract()) {
        findExistingEntry(d_all_abstract_methods, method, true);
      } else {
        findExistingEntry(d_all_abstract_methods, method, false);
      }

      /* 
       * Now process the lists of local methods, provided the method does
       * not belong to a parent.
       */
      if (!isParent) {
        findExistingEntry(d_local_methods, method, true);
        d_local_methods_long.put(long_name, method);
        addToShortMethodList(d_local_methods_short, method);
        if (is_static) {
          findExistingEntry(d_local_static_methods, method, true);
        } else {
          findExistingEntry(d_local_nonstatic_methods, method, true);
        }
      }
    }
  }

  /**
   * Remove the specified method from all method lists.  This is used in
   * From clauses
   *
   * @param  method    The method to be added
   */
  private void removeFromMethodLists(Method method) {
    String long_name = method.getLongMethodName();

    if (!(hasMethodByLongName(long_name, true) && method.isAbstract())) {
      boolean is_static  = method.isStatic();

      /* 
       * First process the lists of local and parent methods (i.e., the
       * "all" lists).
       */
      findExistingEntry(d_all_methods, method, false);
      d_all_methods_long.remove(long_name);
      removeFromShortMethodList(d_all_methods_short, method);
      if (is_static) {
        findExistingEntry(d_all_static_methods, method, false);
      } else {
        findExistingEntry(d_all_nonstatic_methods, method, false);
      }
      findExistingEntry(d_all_abstract_methods, method, false);

      /* 
       * Now process the lists of local methods
       */
      findExistingEntry(d_local_methods, method, false);
      d_local_methods_long.remove(long_name);
      removeFromShortMethodList(d_local_methods_short, method);
      if (is_static) {
        findExistingEntry(d_local_static_methods, method, false);
      } else {
        findExistingEntry(d_local_nonstatic_methods, method, false);
        
      }
    }
  }

  /**
   * Add the invariant to this object.  
   *
   * Assumptions:
   * 1) The constraints on the assertion are being enforced elsewhere.
   * 2) Only invariants local to this extendable are going to be added!
   *
   * @param   assertion  The invariant assertion to be added
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                     The exception raised if the assertion cannot be added.
   */
  public void addInvariant(Assertion assertion) throws AssertionException {
    checkFrozen();
    if (!assertion.isInvariant()) {
      throw new AssertionException("Extendable: " + getFullName(), "Cannot add "
                + "a(n) \"" + assertion.getTypeName() + "\" assertion to the "
                + "invariants list.");
    } else {
      d_invariants.add(assertion);
    } 
  }

  /**
   * Return TRUE if the extendable has any invariants; otherwise, return FALSE.
   *
   * @param   all   If TRUE then check inherited and local; otherwise, check
   *                only local invariants.
   */
  public boolean hasInvariants(boolean all)  throws CodeGenerationException {
    boolean has = false;
    if (getNumberInvariants(false) > 0) {
      has = true;
    } else {
      has = getNumberInvariants(all) > 0 ? true : false;
    }
    return has;
  }

  /**
   * Return the number of invariants in this interface.
   *
   * @param   all   If TRUE then return the number of inherited and local
   *                invariants; otherwise, return only the number of local ones.
   */
  public int getNumberInvariants(boolean all) throws CodeGenerationException {
    return all ? d_invariants.size() : getAllInvariants().size();
  }

  /**
   * Return the invariants in this interface as an <code>List</code>,
   * each element of which is in <code>Assertion</code> form.
   */
  public List getInvariants() {
    return d_invariants;
  }

  /**
   * Return the list of all invariants associated with this object as an 
   * <code>List</code>, each element of which is in <code>Assertion</code>
   * form.
   */
  public List getAllInvariants() throws CodeGenerationException {
    LinkedList  list    = new LinkedList();
    List  parents = getAllParentsInOrder();
    Iterator i = parents.iterator();
    while (i.hasNext()) {
      Extendable parent = (Extendable)i.next();
      list.addAll(parent.getAllInvariants());
    }
    list.addAll(d_invariants);
    return list;
  }

  /**
   * Validate the invariants associated with this object.
   *
   * @param   all   If TRUE then all invariants are to be validated; otherwise,
   *                will only validate the local invariants.
   * @param   skip  If TRUE then will skip the validation process if the
   *                expression has already been marked as valid.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                The exception raised if errors are detected.
   */
  public void validateInvariants(boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    List list = (all ? getAllInvariants() : d_invariants);
    Iterator i = list.iterator();
    while (i.hasNext()) {
      Assertion inv = (Assertion)i.next();
      inv.validateExpression(this, null, skip);
    }
  }

  /**
   * Return TRUE if the extendable has any assertions; otherwise, return FALSE.
   */
  public boolean hasAssertions()  throws CodeGenerationException {
    boolean has = false;
    if (hasInvariants(true)) {
      has = true;
    } else {
      for (Iterator i = getMethods(true).iterator(); i.hasNext() && (!has); ) {
        Method method = (Method) i.next();
        String lname  = method.getLongMethodName();
        if (  (getNumberPreconditions(lname)>0 ) 
           || (getNumberPostconditions(lname)>0) ) { 
          has = true;
        }
      }
    }
    return has;
  }

  /**
   * Return the number of assertions associated with the method in this
   * object, if it is present.  Will return an 0 if there is no such method
   * in this object's hierarchy or if there are no assertions.  Since the
   * invariants are invoked along with the preconditions and again with the
   * postconditions, their number counts twice.
   *
   * @param  longname  The long name of the method the number of whose 
   *                   assertions is to be returned.
   */
  public int getNumberAssertions(String longname)  throws CodeGenerationException {
    return getNumberPreconditions(longname) 
         + 2 * getNumberInvariants(true)
         + getNumberPostconditions(longname);
  }

  /**
   * Return the number of preconditions associated with the method in this
   * object, if it is present.  Will return an 0 if there is no such method
   * in this object's hierarchy or if there are no preconditions.
   *
   * @param  longname  The long name of the method the number of whose 
   *                   preconditions is to be returned.
   */
  public int getNumberPreconditions(String longname) 
    throws CodeGenerationException {
    return getAllPreconditions(longname).size();
  }

  /**
   * Return all preconditions associated with the method in this object,
   * if there is one, in a <code>List</code> with each entry being
   * in <code>Assertion</code> form.  Will return an empty list if no such 
   * method is in this object's hierarchy.
   *
   * WARNING:  
   *   There is no attempt at this point to optimize the preconditions.
   *   That is, duplicate preconditions are NOT removed.
   *
   * @param  longname  The long name of the method whose preconditions are to 
   *                   be returnede
   */
  public List getAllPreconditions(String longname) 
    throws CodeGenerationException {
    LinkedList list    = new LinkedList();
    List parents       = getAllParentsInOrder();
    Iterator  i        = parents.iterator();
    while (i.hasNext()) {
      Extendable parent = (Extendable)i.next();
      list.addAll(parent.getAllPreconditions(longname));
    }
    Method meth = lookupMethodByLongName(longname, false);
    if (meth != null) {
      list.addAll(meth.getPreconditions());
    }
    return list;
  }

  /**
   * Validate the preconditions associated with this object and the specified 
   * method.  By virtue of using getAllPreconditions(), it starts with the 
   * top-most inherited preconditions and proceeds "down" the inheritance 
   * hierarchy to the level of this class.
   *
   * @param   meth  The method whose preconditions are to be validated.
   * @param   all   If TRUE then will validate all preconditions; otherwise,
   *                will validate only the locally defined preconditions.
   * @param   skip  If TRUE then will skip the validation process if the
   *                expression has already been marked as valid.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                The exception raised if errors are detected.
   */
  public void validatePreconditions(Method meth, boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    if (meth != null) {
      List list;
      if (all) {
        String longname = meth.getLongMethodName();
        list = getAllPreconditions(longname);
      } else {
        list = meth.getPreconditions();
      }
      Iterator i = list.iterator();
      while (i.hasNext()) {
        Assertion pre = (Assertion)i.next();
        pre.validateExpression(this, meth, skip);
      }
    } else {
      throw new AssertionException("Extendable: " + getFullName(), 
                  "Cannot validate preconditions for specified method "
                + "when no method given.");
    }
  }

  /**
   * Validate the preconditions associated with this object and the specified 
   * method.  
   *
   * @param   longname  The long name of the method whose preconditions are to 
   *                    be validated.
   * @param   all       If TRUE then will validate all preconditions; otherwise,
   *                    will validate only the locally defined preconditions.
   * @param   skip      If TRUE then will skip the validation process if the
   *                    expression has already been marked as valid.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                    The exception raised if errors are detected.
   */
  public void validatePreconditions(String longname, boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    validatePreconditions(lookupMethodByLongName(longname, true), all, skip);
  }

  /**
   * Validate the preconditions associated with this object.
   *
   * @param   skip  If TRUE then will skip the validation process if the
   *                expression has already been marked as valid.
   * @param   all   If TRUE then will validate all preconditions; otherwise,
   *                will validate only the locally defined preconditions.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                The exception raised if errors are detected.
   */
  public void validatePreconditions(boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    Iterator im;
    if (all) {
      im = d_all_methods.iterator();
    } else {
      im = d_local_methods.iterator();
    }
    while (im.hasNext()) {
      Method meth = (Method)im.next();
      validatePreconditions(meth, all, skip);
    }
  }

  /**
   * Return the number of postconditions associated with the method in this
   * object, if it is present.  Will return an 0 if there is no such method
   * in this object's hierarchy or if there are no postconditions.
   *
   * @param  longname  The long name of the method the number of whose 
   *                   postconditions is to be returned.
   */
  public int getNumberPostconditions(String longname) throws CodeGenerationException {
    return getAllPostconditions(longname).size();
  }

  /**
   * Return all postconditions associated with the method in this object,
   * if there is one, in an <code>List</code> with each element in
   * <code>Assertion</code> form.  Will return an empty list if no such 
   * method in this object's hierarchy.
   *
   * WARNING:  
   *   There is no attempt at this point to optimize the postconditions.
   *   That is, duplicate postconditions are NOT removed.
   *
   * @param  longname  The long name of the method whose postconditions are to 
   *                   be returned.
   */
  public List getAllPostconditions(String longname) 
    throws CodeGenerationException {
    LinkedList  list    = new LinkedList();
    List  parents = getAllParentsInOrder();
    Iterator   i       = parents.iterator();
    while (i.hasNext()) {
      Extendable parent = (Extendable)i.next();
      list.addAll(parent.getAllPostconditions(longname));
    }
    Method meth = lookupMethodByLongName(longname, false);
    if (meth != null) {
      list.addAll(meth.getPostconditions());
    }
    return list;
  }

  /**
   * Validate the postconditions associated with this object and the specified 
   * method.  By virtue of using getAllPostconditions(), it starts with the 
   * top-most inherited postconditions and proceeds "down" the inheritance 
   * hierarchy to the level of this class.
   *
   * @param   meth  The method whose postconditions are to be validated.
   * @param   all   If TRUE then all postconditions of the method will be
   *                validated; otherwise, only the locally defined ones will.
   * @param   skip  If TRUE then will skip the validation process if the
   *                expression has already been marked as valid.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                The exception raised if errors are detected.
   */
  public void validatePostconditions(Method meth, boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    if (meth != null) {
      List list;
      if (all) {
        String longname = meth.getLongMethodName();
        list = getAllPostconditions(longname);
      } else {
        list = meth.getPostconditions();
      }
      Iterator i = list.iterator();
      while (i.hasNext()) {
        Assertion post = (Assertion)i.next();
        post.validateExpression(this, meth, skip);
      }
    } else {
      throw new AssertionException("Extendable: " + getFullName(), 
                  "Cannot validate postconditions for specified method "
                + "when no method given.");
    }
  }

  /**
   * This function returns the set of symbols that this Extendable is
   * dependent on.  It should include any classes or interfaces that are
   * passed by any method in the Extendable, but no enums, and no self.
   */
  public Set getObjectDependencies() throws CodeGenerationException {
    HashSet ret = new HashSet();
    if(d_all_references.size() > 1) {
      for (Iterator i = d_all_references.iterator(); i.hasNext(); ) {
        SymbolID l_id = (SymbolID) i.next();
        Symbol l_sym  = (Symbol) Utilities.lookupSymbol(d_context, l_id);
        //if(l_id != this.getSymbolID() && l_sym.getSymbolType() != Symbol.ENUM) {
        if((l_sym.getSymbolType() != Symbol.ENUM) &&
           (l_sym.getSymbolType() != Symbol.STRUCT)) {
          ret.add(l_id);
        }
      }
    }
    return ret;
  }


  /**
   * Validate the postconditions associated with this object and the 
   * specified method.  
   *
   * @param   longname  The long name of the method whose postconditions are to 
   *                    be validated.
   * @param   all       If TRUE then all postconditions of the method will be
   *                    validated; otherwise, only the locally defined ones 
   *                    will.
   * @param   skip      If TRUE then will skip the validation process if the
   *                    expression has already been marked as valid.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                    The exception raised if errors are detected.
   */
  public void validatePostconditions(String longname, boolean all, boolean skip)
    throws AssertionException, CodeGenerationException
  {
    validatePostconditions(lookupMethodByLongName(longname, true), all, skip);
  }

  /**
   * Validate postconditions associated with this object.
   *
   * @param   skip  If TRUE then will skip the validation process if the
   *                expression has already been marked as valid.
   * @param   all   If TRUE then all postconditions of the method will be
   *                validated; otherwise, only the locally defined ones will.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                The exception raised if errors are detected.
   */
  public void validatePostconditions(boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    Iterator im;
    if (all) {
      im = d_all_methods.iterator();
    } else {
      im = d_local_methods.iterator();
    }
    while (im.hasNext()) {
      Method meth = (Method)im.next();
      validatePostconditions(meth, all, skip);
    }
  }

  /**
   * Validate assertions associated with this object.  That is, validate all 
   * invariants as well as any pre- and post- conditions associated with any 
   * methods.
   *
   * @param   skip  If TRUE then will skip the validation process if the
   *                expression has already been marked as valid.
   * @param   all   If TRUE then all assertions including those for all 
   *                associated methods will be validated; otherwise, only 
   *                validate the locally defined ones.
   * @throws  gov.llnl.babel.symbols.AssertionException
   *                The exception raised if errors are detected.
   */
  public void validateAssertions(boolean all, boolean skip) 
    throws AssertionException, CodeGenerationException
  {
    validateInvariants(all, skip);
    Iterator im;
    if (all) {
      im = d_all_methods.iterator();
    } else {
      im = d_local_methods.iterator();
    }
    while (im.hasNext()) {
      Method meth = (Method)im.next();
      validatePreconditions(meth, all, skip);
      validatePostconditions(meth, all, skip);
    }
  }

  /**
   * Returns true if the specified reserved method call is present in any
   * of the assertions; otherwise, returns false.
   */
  public boolean hasReservedMethodAssertion(int type) throws CodeGenerationException {
    boolean   hasIt = false;
    List list  = getAllInvariants();
    Iterator i = list.iterator();
    while (i.hasNext() && !hasIt) {
      Assertion inv = (Assertion)i.next();
      if (inv.hasReservedMethod(type)) {
        hasIt = true;
      }
    }

    if (!hasIt) {
      Iterator im = d_all_methods.iterator();
      while (im.hasNext() && !hasIt) {
        Method meth = (Method)im.next();
        if (meth.hasReservedMethodAssertion(type)) {
          hasIt = true;
        }
      }
    }
    return hasIt;
  }


  public void freeze()
  {
    if (!d_frozen) {
      super.freeze();
      Iterator i = d_all_methods.iterator();
      while (i.hasNext()) {
        Method m = (Method)i.next();
        m.freeze();
      }
      i = d_all_references.iterator();
      while (i.hasNext()) {
        SymbolID id = (SymbolID)i.next();
        id.freeze();
      }
      i = d_all_basicarrays.iterator();
      while (i.hasNext()) {
        SymbolID id = (SymbolID)i.next();
        id.freeze();
      }
      i = d_renamed_method_symbolID.values().iterator();
      while (i.hasNext()) {
        SymbolID id = (SymbolID)i.next();
        id.freeze();
      }
      d_all_abstract_methods = protectList(d_all_abstract_methods);
      d_all_basicarrays = protectSet(d_all_basicarrays);
      d_all_interfaces = protectMap(d_all_interfaces);
      d_all_methods = protectList(d_all_methods);
      d_all_methods_with_nonblocking = protectList(d_all_methods_with_nonblocking);
      d_all_methods_long = protectMap(d_all_methods_long);
      d_all_methods_short = protectMap(d_all_methods_short);
      d_all_references = protectSet(d_all_references);
      d_all_nonstatic_methods = protectList(d_all_nonstatic_methods);
      d_all_static_methods = protectList(d_all_static_methods);
      d_local_interfaces = protectMap(d_local_interfaces);
      d_local_methods = protectList(d_local_methods);
      d_local_methods_with_nonblocking = protectList(d_local_methods_with_nonblocking);
      d_local_methods_long = protectMap(d_local_methods_long);
      d_local_methods_short = protectMap(d_local_methods_short);
      d_local_nonstatic_methods = protectList(d_local_nonstatic_methods);
      d_local_static_methods = protectList(d_local_static_methods);
      d_renamed_parent_methods = protectMap(d_renamed_parent_methods);
      d_renamed_method_symbolID = protectMap(d_renamed_method_symbolID);
      d_old_name_new_method = protectMap(d_old_name_new_method);
      d_invariants = protectList(d_invariants);
    }
  }
}
